<?php

/**
 * Stará se o uživatele
 */
class User
{
    
    /**
     * @var array Informace o uživateli 
     */
    public $user;
    /**
     * @var array Zprávy k vypsání 
     */
    private $messages;
    
    /**
     * Zjistí, jak je na tom uživatel
     */
    public function __construct()
    {
        // Načte základní funkce a wrapper pro práci s databází
        require 'helpers/main.php';
        require 'Db.php';
        Db::connect();
        
        // Pokud je uložen token, přihlásí to uživatele
        $this->loadUserByToken();
        // Načte uživatele do lokální proměnné pro lehčí prácí
        if (isset($_SESSION['user']))
            $this->user = $_SESSION['user'];
    }
    
    /**
     * Přihlásí uživatele pomocí uloženého tokenu
     */
    private function loadUserByToken()
    {
        if (isset($_COOKIE['Login']) && !$_SESSION['user'])
        {
            $user = Db::queryOne('
                SELECT *
                FROM `user`
                WHERE `token` = ?
            ', array($_COOKIE['Login']));
            
            $_SESSION['user'] = $user;
            // Aktualizuje stránku
            redirect();
        }
    }
    
    /**
     * Přihlásí uživatele
     * @param array $data Data od uživatele
     * @return string V případě selhání vrací chybové hlášky
     */
    public function login($data)
    {
        // Zda-li takový uživatel existuje
        $user = Db::queryOne('
            SELECT *
            FROM `user`
            WHERE `password` = ? AND `username` = ?
        ', array($this->hash($data['password']), $data['username']));
        
        /*
         * Zkontrolování vyplněných políček a jejich správnosti. Také zkontroluje,
         * zda byl účet aktivován
         */
        if (!$data['username'] || !$data['password'])
            $this->addWarning('Nevyplnili jste jméno nebo heslo.');
        if (mb_strlen($data['username']) < 3 || mb_strlen($data['username']) > 255)
            $this->addWarning('Zadejte uživatelské jméno v rozmezí délky 3-255 znaků.');
        if (mb_strlen($data['password']) < 6 || mb_strlen($data['password']) > 255)
            $this->addWarning('Zadejte heslo v rozmezí délky 6-255 znaků.');
        if (!$user[0])
            $this->addWarning('Nesprávné uživatelské jméno nebo heslo.');
        if (!$user['active'] && $user['username'])
            $this->addWarning('Tento účet nebyl ještě aktivován.');
        
        // Pokud neexistují žádné zprávy, přihlásíme uživatele
        if (!$this->getWarnings())
        {
            // Nastavení cookie kvůli trvalému přihlášení na jeden rok dopředu
            if ($data['permanently'])
                setcookie('Login', $user['token'], time() + 31536000, "/");
            
            $_SESSION['user'] = $user;
            redirect('administration');
        }
        else
            return $this->getWarnings();
    }
    
    /**
     * Zaregistruje uživatele
     * @param array $data Data z formuláře
     * @return string Chybové hlášky
     */
    public function registration($data)
    {
        // Zda je použité jméno
        $usedName = Db::querySingle('
            SELECT `user_id`
            FROM `user`
            WHERE `username` = ?
        ', array($data['username']));
        // Zda je použitý E-mail
        $usedEmail = Db::querySingle('
            SELECT `user_id`
            FROM `user`
            WHERE `email` = ?
        ', array($data['email']));
        
        // Chybové hlášky
        if (!$data['username'] || !$data['password'] || !$data['repeat'] || !$data['email'])
            $this->addWarning('Nevyplnili jste všechna políčka.');
        if (mb_strlen($data['username']) < 3 || mb_strlen($data['username']) > 255)
            $this->addWarning('Zadejte uživatelské jméno v rozmezí délky 3-255 znaků.');
        if (mb_strlen($data['password']) < 6 || mb_strlen($data['password']) > 255)
            $this->addWarning('Zadejte heslo v rozmezí délky 6-255 znaků.');
        if ($data['password'] != $data['repeat'])
            $this->addWarning('Hesla se neshodují.');
        if (!preg_match('/[a-zA-Z.0-9\-_]+@[a-zA-Z0-9]+\.[a-z]{2,6}/', $data['email']))
            $this->addWarning('E-mail není ve správném formátu.');
        if ($usedName)
            $this->addWarning('Toto uživatelské jméno je již používáno.');
        if ($usedEmail)
            $this->addWarning('Tento E-mail je již obsazen.');
        
        // Pokud neexistují žádné zprávy, můžeme uživatele zaregistrovat
        if (!$this->getWarnings())
        {
            $password = $this->hash($data['password']);
            $token = $this->hash($data['username']);
            
            // Data k registraci uživatele
            $insert = array(
                'username' => $data['username'],
                'password' => $password,
                'email' => $data['email'],
                'token' => $token,
                'active' => 0
            );
            Db::insert('user', $insert);
            
            // Data k aktivaci účtu
            $insert = array(
                'user' => Db::getLastId(),
                'token' => $token
            );
            Db::insert('active_user', $insert);
            
            $this->sendActivationEmail($token, $data['email']);
            // Přidá zprávu o aktivaci účtu
            $this->addWarning('Pro úplné zaregistrování účtu je nutné ho aktivovat '
                    . 'přes aktivační odkaz, který byl odeslán na Váš E-mail.');
            // Vyprázdnění formuláře
            unset($_POST);
        }
        return $this->getWarnings();
    }
    
    /**
     * Aktivuje účet
     * @param string $token Token z E-mailu
     * @return string Zpráva pro uživatele
     */
    public function activeUser($token)
    {
        $user = Db::querySingle('
            SELECT `user`
            FROM `active_user`
            WHERE `token` = ?
        ', array($token));
        
        // Pokud tento token existuje
        if ($user)
        {
            // Odstraní záznam, aby se zbytečně nedal účet aktivovat několikrát
            Db::query('
                DELETE
                FROM `active_user`
                WHERE `user` = ?
            ', array($user));
            // Aktivuje účet
            Db::update('user', array('active' => 1), 'WHERE `user_id` = ?', array($user));
            
            return 'Účet byl úspěšně aktivován.';
        }
        
        return 'Tento token neexistuje.';
    }
    
    /**
     * Vygeneruje nové heslo a odešle ho na E-mail
     * @param array $data Data z formuláře
     * @return string Chybové hlášky
     */
    public function newPassword($data)
    {
        // Kontrola existence uživatele
        $existUser = Db::querySingle('
            SELECT `user_id`
            FROM `user`
            WHERE `username` = ? AND `email` = ?
        ', array($data['username'], $data['email']));
        
        // Kontrola formuláře
        if (!$data['username'] || !$data['email'])
            $this->addWarning('Nevyplnili jste všechna políčka.');
        if (!preg_match('/[a-zA-Z.0-9\-_]+@[a-zA-Z0-9]+\.[a-z]{2,6}/', $data['email']))
            $this->addWarning('E-mail není ve správném formátu.');
        if (!$existUser)
            $this->addWarning('Uživatel s tímto jménem nebo E-mailem neexistuje.');
        
        if (!$this->getWarnings())
        {
            // Zkontrolování aktivovaného účtu
            $active = Db::querySingle('
                SELECT `active`
                FROM `user`
                WHERE `username` = ?
            ', array($data['username']));
            
            if (!$active)
            {
                $this->addWarning('Tento účet nebyl ještě aktivován, tudíž nelze '
                        . 'změnit heslo. Nejdříve tedy aktivujte Váš účet.');
            }
            // Aktivovaný účet
            else
            {
                // Nově vygenerované heslo
                $password = $this->hash(randomString(rand(10, 20)));

                Db::update('user', array('password' => $password),
                        'WHERE `username` = ?', array($data['username']));
                $this->sendPasswordEmail($data['email'], $password);
                $this->addWarning('Nové heslo Vám bylo odesláno na Váš E-mail.');
                
                unset($_POST);
            }
        }
        return $this->getWarnings();
    }
    
    /**
     * Odhlásí uživatele
     */
    public function logout()
    {
        unset($_SESSION['user']);
        setcookie('Login', '', time() - 3600, "/");
        redirect('login');
    }
    
    /**
     * Pokud je (ne)přihlášený člověk na špatné stránce, přesměruje ho to na
     * tu správnou stránku
     * @param bool $administration Zda se jedná o administraci
     */
    public function checkPage($administration)
    {
        if (!$administration && $this->user)
            redirect('administration');
        elseif ($administration && !$this->user)
            redirect('login');
    }
    
    /**
     * Odešle aktivační E-mail
     * @param string $token Token k aktivaci
     * @param string $address Adresát
     */
    private function sendActivationEmail($token, $address)
    {
        require 'helpers/email.php';
        $message = 'Milý uživateli, milá uživatelko,<br />'
                . 'registrovali jste se na síť <a href="http://machr">Machr</a>. '
                . '<a href="http://machr/activation?token=' . $token
                . '">Aktivujte si účet a získejte plno výhod!</a><br />'
                . 'Pokud jste se na tyto webové stránky nezaregistrovali, ignorujte '
                . 'tento E-mail.<br />S pozdravem<br />IT Man';
        sendEmail($address, 'Aktivace účtu', $message, 'admin@machr.cz');
    }
    
    /**
     * Odešle E-mail s novým heslem
     * @param string $address Adresát
     * @param string $password Nové heslo
     */
    private function sendPasswordEmail($address, $password)
    {
        require 'helpers/email.php';
        $message = 'Milý uživateli, milá uživatelko,<br />'
                . 'nechali jste si ze sítě <a href="http://machr">Machr</a> odeslat '
                . 'nové heslo. Pokud jste si ho neodesílali Vy, Váš účet je zřejmě v '
                . 'ohrožení a doporučujeme změnu hesla.<br />'
                . 'Nové heslo je: <strong>' . $password . '</strong><br />'
                . 'Děkujeme za používání našich webových stránek.<br />'
                . 'S pozdravem<br />IT Man';
        sendEmail($address, 'Nové heslo', $message, 'admin@machr.cz');
    }
    
    /**
     * Vrací zahashované heslo
     * @param string $password Normální heslo
     * @return string Zahashované heslo
     */
    private function hash($password)
    {
        return hash('sha512', md5($password . 'f8#4as+1me'));
    }
    
    /**
     * Přidá zprávu pro uživatele
     * @param string $message Zpráva
     */
    private function addWarning($message)
    {
        $this->messages[] = $message;
    }
    
    /**
     * Vrací zprávy uživateli
     * @return string Výsledné HTML se zprávami
     */
    private function getWarnings()
    {
        if ($this->messages)
            return implode('<br />', $this->messages);
        return false;
    }
    
}
